package com.elphel.imagej.orthomosaic;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;

import Jama.Matrix;

public class PairwiseOrthoMatch implements Serializable {
	private static final long serialVersionUID = 1L;
	public double [][]           affine = new double[2][3];
	public transient double [][] jtj =    new double [6][6];
	public int zoom_lev;
	public double rms = Double.NaN;
	public transient int [] nxy = null; // not saved, just to communicate for logging
	public transient double overlap = 0.0;
	public PairwiseOrthoMatch() {
		
	}
	public PairwiseOrthoMatch(
			double [][] affine,
			double [][] jtj,
			double rms,
			int zoom_lev,
			double overlap) {
		this.affine =  affine;
		this.jtj =     jtj;
		this.zoom_lev= zoom_lev;
		this.rms =     rms;
		this.overlap = overlap;
	}
	public PairwiseOrthoMatch clone() {
		double [][] affine = {this.affine[0].clone(),this.affine[1].clone()};
		double [][] jtj = new double [this.jtj.length][];
		for (int i = 0; i < this.jtj.length; i++) {
			jtj[i] = this.jtj[i].clone();
		}
		PairwiseOrthoMatch pom = new PairwiseOrthoMatch(
				affine,
				jtj,
				this.rms,
				this.zoom_lev,
				this.overlap);
		if (nxy != null) {
			pom.nxy = nxy.clone();
		}
		return pom;
	}
	/**
	 * If this match is calculated from scene0 to scene1, inverse one is from scene1 to scene0  
	 * @param rd displacement from scene0 reference (vertical) point to scene1 reference point
	 * (GNSS-derived), referenced as -V in the notes below. (it is calculated as 0 from 1)
	 * @return inverted 3x2 matrix
	 * 
		Xs0, Xs1 - metric coordinates (from top-left) in each of the source images
		V - image center offset of the second scene from the first in rectified space (GPS-derived)
		S0, S1 - vertical point offset from the top-left in image coordinates
		A0,A1 - affine transform matrices (sensor coordinates from rectified coordinates)
		B0, B1 centers offsets in image coordinates
		Xr - rectified coordinates. Xr1, Xr2 - just different points
		
		Xsi = Ai*(Xr - V) + Bi + Si
		
		1) A00 = E, A1, B1 - second image match when the first is unity:
		
		Xs0 = Xr1 + S0
		Xs1 = A1 * (Xr1 - V) + B1 + S1;
		
		2) calculate inverted A1,B1 of the first scene relative to the second one. In this case Xs0
		of the first scene should still correspond to Xs1
		Xs0 = A0 * (Xr2 + V) + B0 + S0;
		Xs1 = Xr2 + S1
		
		----
		Xs1 =  A1 * (Xs0 - S0 - V) + B1 + S1;
		Xs1 - B1 - S1 =  A1 * (Xs0 - S0 - V);
		A1inv*(Xs1 - B1 - S1) =  Xs0 - S0 - V;
		Xs0 = A1inv*(Xs1 - B1 - S1) + S0 + V
		---
		Xr2 = Xs1 - S1
		Xs0 = A0 * (Xs1 - S1 + V) + B0 + S0;
		
		A1inv*(Xs1 - B1 - S1) + S0 + V = A0 * (Xs1 - S1 + V) + B0 + S0
		=> A0=A1inv
		A1inv*(Xs1 - B1 - S1) + S0 + V = A1inv * (Xs1 - S1 + V) + B0 + S0
		-A1inv*( B1)  + V = A1inv * ( V) + B0
		A1inv * ( V) + B0 + A1inv*(B1) - V = 0
		B0 = -((A1inv-E)*V + A1inv*B1)
	 */
	
	public PairwiseOrthoMatch getInverse(double [] rd) {
		double [][] affine = OrthoMap.invertAffine(getAffine());
		PairwiseOrthoMatch inverted_match = new PairwiseOrthoMatch(
				affine, // double [][] affine,
				jtj, // double [][] jtj,
				rms, // double rms,
				zoom_lev, // int zoom_lev)
				overlap); // 
		double [] corr = {
				rd[0] * (affine[0][0]-1.0)+ rd[1]*affine[0][1],
				rd[0] * affine[1][0]+       rd[1]*(affine[1][1]-1.0)};
		affine[0][2] += corr[0];
		affine[1][2] += corr[1];
		return inverted_match;	
	}
	
	/**
	 * Create differential PairwiseOrthoMatch instance from 2 affines of the scenes and an offset vector
	 * @param affine0
	 * @param affine1
	 * @param rd
	 * Continued from getInverse notes:
	 * =========================== V = -rd
		Xsi = Ai*(Xr - V) + Bi + Si
		
		Xs0 = A0*(Xr - 0) + B0 + S0
		Xs1 = A1*(Xr - V) + B1 + S1
		
		--- difference Ad, Bd 
		Xs0 = Xr1 + S0
		Xs1 = Ad * (Xr1 - V) + Bd + S1;
		
		
		Xs0 = A0*(Xr - 0) + B0 + S0
		Xs0 - (B0 + S0) = A0*Xr
		Xr = A0inv*(Xs0 - (B0 + S0))
		Xs1 = A1*((A0inv*(Xs0 - (B0 + S0))) - V) + B1 + S1
		
		Xr1 = Xs0-S0
		
		Xs1 = Ad * ((Xs0-S0) - V) + Bd + S1;
		
		A1*((A0inv*(Xs0 - (B0 + S0))) - V) + B1 + S1 = Ad * ((Xs0-S0) - V) + Bd + S1
		
		A1*((A0inv*(Xs0 - (B0 + S0))) - V) + B1 = Ad * ((Xs0-S0) - V) + Bd
		
		A1*((A0inv*(Xs0 - (B0 + S0))) - V) + B1 = Ad * ((Xs0-S0) - V) + Bd
		A1 * A0inv * Xs0 - A1 * A0inv * (B0 + S0) - A1*V + B1 - Ad *Xs0 + Ad * S0 +Ad*V -Bd
		=> Ad = A1 * A0inv:
		 - A1 * A0inv * (B0 + S0) - A1*V + B1  + Ad * S0 +Ad*V -Bd =
		 - Ad * (B0 + S0) - A1*V + B1  + Ad * S0 + Ad*V - Bd =
		 Ad * (S0 + V - B0 - S0) - A1*V + B1- Bd = 
		  Ad * (V - B0) - A1 * V + B1 - Bd
		Bd =  Ad * (V - B0) - A1 * V + B1
		
		--- check if A0 = E, B0 = 0
		Ad = A1inv
		Bd = A1 * V - A1 *V + B1 = B1
		--- check if A1 = E, B1 = 0, V1= -V
		Bd =  A0inv * (V1 - B0) - V1 =
		(A0inv-E)*V1 -A0inv*B0 = 
		-(A0inv-E)*V -A0inv*B0 = 
	 */
	public PairwiseOrthoMatch (
			double [][] affine0,
			double [][] affine1,
			double [] rd) {
		Matrix A0 = new Matrix (
				new double [][] {{affine0[0][0],affine0[0][1]},{affine0[1][0],affine0[1][1]}});
		Matrix B0 = new Matrix(new double [][] {{affine0[0][2]},{affine0[1][2]}});
		Matrix A1 = new Matrix (
				new double [][] {{affine1[0][0],affine1[0][1]},{affine1[1][0],affine1[1][1]}});
		Matrix B1 = new Matrix(new double [][] {{affine1[0][2]},{affine1[1][2]}});
		Matrix V =  new Matrix(new double [][] {{-rd[0]},{-rd[1]}});
		Matrix A =  A1.times(A0.inverse());
		Matrix B = A.times(V.minus(B0)).minus(A1.times(V)).plus(B1);
		affine = new double[][] {
				{A.get(0,0),A.get(0,1), B.get(0,0)},
				{A.get(1,0),A.get(1,1), B.get(1,0)}};
//				jtj = null;
//				rms = Double.NaN; // double rms,
//				zoom_lev = 0; // int zoom_lev)
	}
	
	public double [][] getAffine(){
		return affine;
	}
	
	private void writeObject(ObjectOutputStream oos) throws IOException {
		oos.defaultWriteObject();
		for (int i = 0; i < jtj.length; i++) {
			for (int j = i; j < jtj[i].length; j++) {
				oos.writeObject(jtj[i][j]);
			}
		}
		oos.writeObject(overlap);
	}
	private void readObject(ObjectInputStream ois) throws ClassNotFoundException, IOException {
		ois.defaultReadObject();
		jtj = new double [6][6]; // readObject does not use constructor!
		for (int i = 0; i < jtj.length; i++) {
			for (int j = i; j < jtj[i].length; j++) {
				jtj[i][j] = (Double) ois.readObject();
				if (j > i) {
					jtj[j][i] = jtj[i][j];
				}
			}
		}
		overlap = (Double) ois.readObject();
	}
	//private void readObjectNoData() throws ObjectStreamException; // used to modify default values
}
